################################################################################
#  xregistry.rb
#
#  This file defines a cross-platform Registry class for storing application
#  configuration parameters.
#
#  The application programming interface (API) is based on the FOX registry
#  class, which is part of the FOX GUI toolkit (http://www.fox-toolkit.org/).
#  This file defines a class that can be used in the same way as the FOX
#  registry, but without needing the FOX toolkit.
#
#  The original FOX registry uses the Windows registry if running under Windows,
#  and an .INI type file if running under Unix-type systems.  This class also
#  uses the Windows registry if running under Windows, but uses an SQLite3
#  database if running under a Unix-type system.
#
#  Last revised:  03-Aug-2016  Wm. Parsons
################################################################################


class Registry
  # Construct registry object; app_key and vendor_key must be string constants.
  # Regular applications SHOULD set a vendor key!
  def initialize(app_key = '', vendor_key = '')
    @app_key    = app_key
    @vendor_key = vendor_key
    @on_windows = RUBY_PLATFORM =~ /w(in)?32/

    if @on_windows
      require 'win32/registry'
      @root_key = ['SOFTWARE', vendor_key, app_key].join('\\')
    else
      require 'sqlite3'

      # create registry directory if necessary
      regdir = ENV['HOME'] + '/.registry'
      File.directory?(regdir) || Dir.mkdir(regdir)
      regdir += '/' + vendor_key
      File.directory?(regdir) || Dir.mkdir(regdir)

      @db = SQLite3::Database.new("#{regdir}/#{app_key}.db3")
    end
  end

  # Return application key
  def appKey
    @app_key
  end
  alias app_key appKey

  # Return vendor key
  def vendorKey
    @vendor_key
  end
  alias vendor_key vendorKey

  # Delete the registry entry for the specified section and key.
  def deleteEntry(section, key)
    if @on_windows
      begin
        pkey = [@root_key, section].join('\\')
        access = Win32::Registry::KEY_ALL_ACCESS
        Win32::Registry::HKEY_CURRENT_USER.open(pkey, access) do |reg|
          reg.delete_value(key)
        end
      rescue Win32::Registry::Error
      end
    else
      begin
        @db.execute %Q[delete from "#{section}" where key=?], key
      rescue SQLite3::SQLException
      end
    end
  end
  alias delete_entry deleteEntry

  # Delete a specified section from the registry.
  def deleteSection(section)
    if @on_windows
      begin
        access = Win32::Registry::KEY_ALL_ACCESS
        Win32::Registry::HKEY_CURRENT_USER.open(@root_key, access) do |reg|
          reg.delete_key(section, true)
        end
      rescue Win32::Registry::Error
      end
    else
      @db.execute "drop table if exists \"#{section}\""
    end
  end
  alias delete_section deleteSection

  # Iterate over the section keys of the registry.
  def each_key
    if @on_windows
      Win32::Registry::HKEY_CURRENT_USER.open(@root_key) do |reg|
        reg.each_key { |subkey, wtime| yield subkey }
      end
    else
      @db.execute <<-EOS do |name|
          select name from sqlite_master where type='table' order by name
          EOS
        yield name[0]
      end
    end
  end

  # Find a section in the registry.
  # The properties of the value returned are undefined except that it must
  # support an "each_key" to iterate thru the keys and a "find" method to
  # locate the value associated with a key.
  def find(section)
    hash = {}
    if @on_windows
      pkey = [@root_key, section].join('\\')
      Win32::Registry::HKEY_CURRENT_USER.open(pkey) do |reg|
        reg.each do |subkey, type, value|
          hash[subkey] = value
        end
      end
    else
      @db.execute %Q[select key, value from "#{section}" order by key] do |row|
        hash[row[0]] = row[1]
      end
    end
    class << hash
      def find(key)
        self[key]
      end
    end
    hash
  end

  # Read a boolean registry entry from the specified section and key.  If no
  # value is found, the default value is returned.  A type error is
  # raised if the retrieved value is real or a string.
  def readBoolEntry(section, key, default = false)
    value = readEntry(section, key, default)
    raise TypeError if value.class == Float || value.class == String
    value && value != 0
  end
  alias read_bool_entry readBoolEntry

  # Read an integer registry entry from the specified section and key.  If no
  # value is found, the default value is returned.
  def readIntEntry(section, key, default = 0)
    readEntry(section, key, default).to_i
  end
  alias read_int_entry readIntEntry

  # Read a double-precision floating point registry entry from the specified
  # section and key.  If no value is found, the default value is returned.
  def readRealEntry(section, key, default = 0.0)
    readEntry(section, key, default).to_f
  end
  alias read_real_entry readRealEntry

  # Read a string registry entry from the specified section and key.  If no
  # value is found, the default value is returned.
  def readStringEntry(section, key, default = '')
    readEntry(section, key, default).to_s
  end
  alias read_string_entry readStringEntry

  # Write a boolean registry value to the specified section and key.
  def writeBoolEntry(section, key, value)
    writeEntry(section, key, value ? 1 : 0)
  end
  alias write_bool_entry writeBoolEntry

  # Write an integer registry value to the specified section and key.
  def writeIntEntry(section, key, value)
    writeEntry(section, key, value.to_i)
  end
  alias write_int_entry writeIntEntry

  # Write a double-precision floating point registry value to the specified
  # section and key.
  def writeRealEntry(section, key, value)
    writeEntry(section, key, value.to_f)
  end
  alias write_real_entry writeRealEntry

  # Write a string registry value to the specified section and key.
  def writeStringEntry(section, key, value)
    writeEntry(section, key, value.to_s)
  end
  alias write_string_entry writeStringEntry

  #######
  private
  #######

  # Read a registry entry from the specified section and key.  If no
  # value is found, the default value is returned.
  def readEntry(section, key, default)
    if @on_windows
      pkey = [@root_key, section].join('\\')
      reg  = Win32::Registry::HKEY_CURRENT_USER.open(pkey)
      reg.read(key)[1]
    else
      @db.get_first_value("select value from \"#{section}\" where key=?",
                          key) || default
    end
  rescue
    default
  end

  # Write a registry value to the specified section and key.
  def writeEntry(section, key, value)
    if @on_windows
      # store real values as strings under Windows
      value = value.to_s if value.class == Float
      begin
        pkey = [@root_key, section].join('\\')
        access = Win32::Registry::KEY_WRITE
        Win32::Registry::HKEY_CURRENT_USER.open(pkey, access) do |reg|
          reg[key] = value
        end
      rescue
        Win32::Registry::HKEY_CURRENT_USER.create(pkey, access) do |reg|
          reg[key] = value
        end
      end
    else
      begin
        @db.execute("update \"#{section}\" set value=? where key=?",
                    [value, key])
        if @db.changes == 0
          # no row was updated so insert a new row
          @db.execute("insert into \"#{section}\"(key, value) values(?,?)",
                      [key, value])
        end
      rescue SQLite3::Exception => e
        if e.to_s =~ /no such table/
          @db.execute "create table \"#{section}\"(key text unique, value)"

          @db.execute("insert into \"#{section}\"(key, value) values(?,?)",
                      [key, value])
        else
          raise e.to_s
        end
      end
    end
  end
end
